@@ -66,10 +66,15 @@
 #include <libintl.h>
 #include <sys/sysmacros.h>
 #include <linux/major.h>
-#include <utmp.h>
 #include <security/pam_appl.h>
 #include <security/pam_misc.h>
 
+#ifdef WITH_SELINUX
+#include <selinux/get_context_list.h>
+#include <selinux/flask.h>
+#include <selinux/selinux.h>
+#endif
+
 #include "faillog.h"
 #include "failure.h"
 
@@ -91,6 +96,14 @@
 #define _(s) gettext(s)
 #endif
 
+#ifdef WITH_SELINUX
+int selinux_enabled=0;
+security_context_t user_context=NULL;
+security_context_t ttyn_context=NULL; /* The current context of ttyn device */
+security_context_t vcsn_context=NULL; /* The current context of vcsn device */
+security_context_t vcsan_context=NULL;/* The current context of vcsan device */
+#endif
+
 /* Print the version information.  */
 static inline void
 print_version (void)
@@ -199,6 +212,53 @@ timedout (int sig __attribute__ ((unused)))
   exit (0);
 }
 
+#ifdef WITH_SELINUX
+
+static security_context_t
+security_label_tty (const char *tty, security_context_t usercon)
+{
+  security_context_t newdev_context=NULL; /* The new context of a device */
+  security_context_t prev_context=NULL; /* The new context of a device */
+  if (getfilecon (tty, &prev_context) < 0)
+    {
+      fprintf (stderr, _("Warning!  Could not get current context for %s, not relabeling.\n"), tty);
+      syslog (LOG_NOTICE, "Warning!  Could not get current context for %s, notrelabeling.\n", tty);
+      return NULL;
+    }
+  if (security_compute_relabel (usercon, prev_context, SECCLASS_CHR_FILE,
+				&newdev_context) != 0)
+    {
+      fprintf (stderr, _("Warning!  Could not get new context for %s, not relabeling.\n"), tty);
+      syslog (LOG_NOTICE, "Warning!  Could not get new context for %s, not relabeling.\n", tty);
+      return NULL;
+    }
+  if (setfilecon (tty,newdev_context))
+    {
+      fprintf (stderr, _("Warning!  Could not relabel %s with %s, not relabeling.\n"), tty, newdev_context);
+      syslog (LOG_NOTICE, "Warning!  Could not relabel %s with %s, not relabeling.\n", tty, newdev_context);
+      return NULL;
+    }
+  freecon (newdev_context);
+  return prev_context;
+}
+
+static void
+security_restorelabel_tty (const char *tty, security_context_t context)
+{
+  if (context == NULL)
+    return;
+
+  if (setfilecon (tty, context))
+    {
+      fprintf (stderr, _("Warning!  Could not relabel %s with %s, not relabeling.\n"), tty, context);
+      syslog (LOG_NOTICE, "Warning!  Could not relabel %s with %s, not relabeling.\n", tty, context);
+    }
+  freecon (context);
+  return;
+}
+#endif /* WITH_SELINUX */
+
+
 /* Nice and simple code provided by Linus Torvalds 16-Feb-93 */
 /* Nonblocking stuff by Maciej W. Rozycki, macro@ds2.pg.gda.pl, 1999.
    He writes: "Login performs open() on a tty in a blocking mode.
@@ -215,9 +275,9 @@ opentty (const char *tty)
 
   if (fd == -1)
     {
-      syslog (LOG_ERR, _("FATAL: can't reopen tty: %s"),
+      syslog (LOG_ERR, "FATAL: can't reopen tty: %s",
 	      strerror (errno));
-      sleep (1);
+      sleep (5);
       exit (1);
     }
 
@@ -232,9 +292,9 @@ opentty (const char *tty)
       {
 	if (dup2 (fd, i) == -1)
 	  {
-	    syslog (LOG_ERR, _("FATAL: can't reopen stdin/stdout/stderr: %s"),
+	    syslog (LOG_ERR, "FATAL: can't reopen stdin/stdout/stderr: %s",
 		    strerror (errno));
-	    sleep (1);
+	    sleep (5);
 	    exit (1);
 	  }
       }
@@ -257,8 +317,8 @@ check_ttyname (char *ttyn)
       || (statbuf.st_nlink > 1 &&
 	  strncmp(ttyn, _PATH_DEV, strlen (_PATH_DEV)) != 0))
     {
-      syslog (LOG_ERR, _("FATAL: bad tty"));
-      sleep (1);
+      syslog (LOG_ERR, "FATAL: bad tty");
+      sleep (5);
       exit (1);
     }
 }
@@ -469,6 +529,8 @@ main (int argc, char **argv)
   char *buffer = alloca (buflen);
   struct passwd resultbuf;
   struct passwd *pwd;
+  char vcsn[20], vcsan[20];
+
 
   init_sighandler ();
 
@@ -644,11 +706,11 @@ main (int argc, char **argv)
 	  (chmod (ttyn, getdef_num ("TTYPERM", 0600)) == -1))
 	{
 	  errsv = errno;
-	  syslog (LOG_ERR, _("FATAL: cannot change permissions of TTY: %s"),
+	  syslog (LOG_ERR, "FATAL: cannot change permissions of TTY: %s",
 		  strerror (errsv));
 	  fprintf (stderr, _("FATAL: cannot change permissions of TTY: %s"),
 		   strerror (errsv));
-	  sleep (1);
+	  sleep (5);
 	  exit (1);
 	}
 
@@ -795,7 +857,7 @@ main (int argc, char **argv)
   if (username == NULL || *username == '\0')
     {
       fprintf (stderr, _("\nSession setup problem, abort.\n"));
-      syslog (LOG_ERR, _("NULL user name in %s:%d. Abort."), __FUNCTION__, __LINE__);
+      syslog (LOG_ERR, "NULL user name in %s:%d. Abort.", __FUNCTION__, __LINE__);
       pam_end (pamh, PAM_SYSTEM_ERR);
       exit (1);
     }
@@ -811,7 +873,7 @@ main (int argc, char **argv)
 
   if(!pwd)
     {
-      syslog (LOG_ERR, _("Failed to look up user '%s'."), username);
+      syslog (LOG_ERR, "Failed to look up user '%s'.", username);
       pam_end (pamh, PAM_SYSTEM_ERR);
       exit (1);
     }
@@ -858,25 +920,68 @@ main (int argc, char **argv)
   retcode = pam_open_session (pamh, 0);
   PAM_FAIL_CHECK;
 
+#ifdef WITH_SELINUX
+  /* Make sure SELINUX is really running on this system */
+  if ((selinux_enabled = is_selinux_enabled ()))
+    {
+      int ret;
+      security_context_t* contextlist;
+      int num_contexts = get_ordered_context_list (username, 0, &contextlist);
+      if (num_contexts > 0)
+	{
+#if 0
+	  ret = query_user_context (contextlist, &user_context);
+	  freeconary (contextlist);
+
+	  if (ret < 0)
+	    {
+	      fprintf (stderr, _("%s:  query_user_context failed\n"), argv[0]);
+	      syslog (LOG_ERR, "%s:  query_user_context failed\n", argv[0]);
+	      sleep (5);
+	      exit (3);
+	    }
+#else
+	  user_context = (security_context_t) strdup (contextlist[0]);
+	  freeconary (contextlist);
+#endif
+	}
+      else
+	{
+	  fprintf (stderr, _("login: unable to obtain context for %s.\n"),
+		   username);
+	  ret = manual_user_enter_context (username, &user_context);
+	  if (ret < 0)
+	    {
+	      syslog (LOG_ERR, "UNABLE TO GET VALID CONTEXT FOR %s",
+		      username);
+	      sleep (5);
+	      exit (3);
+	    }
+	}
+    }
+#endif
+
   /* committed to login -- turn off timeout */
   alarm ((unsigned int) 0);
 
 #ifdef HAVE_QUOTA
   if (quota (Q_SETUID, pwd->pw_uid, 0, 0) < 0 && errno != EINVAL)
     {
-      switch (errno) {
-      case EUSERS:
-	fprintf(stderr,
-		_("Too many users logged on already.\nTry again later.\n"));
-	break;
-      case EPROCLIM:
-	fprintf(stderr,
-		_("You have too many processes running.\n"));
-	break;
-      default:
-	perror("quota (Q_SETUID)");
-      }
-      sleepexit(0);           /* %% */
+      switch (errno)
+	{
+	case EUSERS:
+	  fprintf (stderr,
+		   _("Too many users logged on already.\nTry again later.\n"));
+	  break;
+	case EPROCLIM:
+	  fprintf (stderr,
+		   _("You have too many processes running.\n"));
+	  break;
+	default:
+	  perror ("quota (Q_SETUID)");
+	}
+      sleep (5);
+      exit (0);           /* %% */
     }
 #endif
 
@@ -1013,13 +1118,15 @@ main (int argc, char **argv)
   chown (ttyn, pwd->pw_uid, gid);
   chmod (ttyn,  getdef_num ("TTYPERM", 0600));
 
+#ifdef WITH_SELINUX
+  if (selinux_enabled)
+    ttyn_context = security_label_tty (ttyn, user_context);
+#endif
+
   /* if tty is one of the VC's then change owner and mode of the
      special /dev/vcs devices as well */
   if (consoletty (0))
     {
-#if 0
-      char vcsn[20], vcsan[20];
-
       /* find names of Virtual Console devices */
       char *p = ttyn;
       /* find number of tty */
@@ -1030,7 +1137,7 @@ main (int argc, char **argv)
       strcat (vcsn, p);
       strcpy (vcsan, "/dev/vcsa");
       strcat (vcsan, p);
-
+#if 0
       /*
        * Please don't add code to chown /dev/vcs* to the user logging in -
        * it's a potential security hole.  I wouldn't like the previous user
@@ -1042,6 +1149,13 @@ main (int argc, char **argv)
       chown (vcsan, pwd->pw_uid, gid);
       chmod (vcsn, getdef_num ("TTYPERM", 0600));
       chmod (vcsan, getdef_num ("TTYPERM", 0600));
+#endif
+#ifdef WITH_SELINUX
+      if (selinux_enabled)
+        {
+	  vcsn_context = security_label_tty (vcsn, user_context);
+	  vcsan_context = security_label_tty (vcsan, user_context);
+	}
 #endif
     }
 
@@ -1123,6 +1237,30 @@ main (int argc, char **argv)
 
   setproctitle ("login", username);
 
+#ifdef WITH_SELINUX
+  if (selinux_enabled)
+    {
+      if (pwd->pw_uid == 0)
+	{
+	  if (hostname)
+	    syslog (LOG_NOTICE, "ROOT LOGIN ON %s FROM %s USING %s",
+		    ttyn, hostname, user_context);
+	  else
+	    syslog (LOG_NOTICE, "ROOT LOGIN ON %s USING %s", ttyn,
+		    user_context);
+	}
+      else
+	{
+	  if (hostname)
+	    syslog (LOG_INFO, "LOGIN ON %s BY %s FROM %s USING %s",
+		    ttyn, pwd->pw_name, hostname, user_context);
+	  else
+	    syslog (LOG_INFO, "LOGIN ON %s BY %s USING %s", ttyn,
+		    pwd->pw_name, user_context);
+	}
+    }
+#endif
+
   if (!quietlog)
     motd ();
 
@@ -1170,6 +1308,20 @@ main (int argc, char **argv)
 	    fprintf (stderr, _("login: waitpid (%d, NULL, 0) failed: %s\n"),
 		     childPid, strerror (errsv));
 	}
+#ifdef WITH_SELINUX
+      if (selinux_enabled)
+        {
+          /* We need to change the contexts of the terminal devices back to
+             the system when the user's session ends.  */
+          security_restorelabel_tty (ttyn, ttyn_context);
+          if (consoletty (0))
+	    {
+	      security_restorelabel_tty (vcsn, vcsn_context);
+	      security_restorelabel_tty (vcsan, vcsan_context);
+	    }
+        }
+#endif
+
       PAM_END;
       exit (0);
     }
@@ -1188,6 +1340,22 @@ main (int argc, char **argv)
 
   openlog ("login", LOG_ODELAY | LOG_PID, LOG_AUTHPRIV);
 
+#ifdef WITH_SELINUX
+  if (selinux_enabled)
+    {
+      if (setexeccon (user_context))
+	{
+	  fprintf (stderr,
+		   _("Error!  Unable to set executable context %s.\n"),
+		   user_context);
+	  syslog(LOG_ERR, "Error!  Unable to set executable context %s.\n",
+		 user_context);
+	  exit (3);
+	}
+      freecon (user_context);
+    }
+#endif
+
   /* discard permissions last so can't get killed and drop core */
   if (setuid (pwd->pw_uid) < 0 && pwd->pw_uid)
     {
